@@ -9,30 +9,27 @@ from sql import Null
 from sql.aggregate import Sum
 from sql.conditionals import Coalesce
 
-from trytond.model import ModelView, Workflow, ModelSQL, fields, Check, Unique
+from trytond.i18n import gettext
+from trytond.model import ModelView, Workflow, ModelSQL, fields, Unique
+from trytond.model.exceptions import AccessError
 from trytond.wizard import Wizard, StateView, StateTransition, Button
 from trytond.pyson import Not, Equal, Eval, Or, Bool, If
 from trytond.transaction import Transaction
 from trytond.pool import Pool
 from trytond.tools import reduce_ids, grouped_slice
 
-__all__ = ['Forecast', 'ForecastLine', 'ForecastLineMove',
-    'ForecastCompleteAsk', 'ForecastCompleteChoose', 'ForecastComplete']
-
-STATES = {
-    'readonly': Not(Equal(Eval('state'), 'draft')),
-}
-DEPENDS = ['state']
-FORECAST_STATES = [
-    ('draft', 'Draft'),
-    ('done', 'Done'),
-    ('cancel', 'Cancel'),
-    ]
+from .exceptions import ForecastValidationError
 
 
 class Forecast(Workflow, ModelSQL, ModelView):
     "Stock Forecast"
     __name__ = "stock.forecast"
+
+    _states = {
+        'readonly': Not(Equal(Eval('state'), 'draft')),
+    }
+    _depends = ['state']
+
     warehouse = fields.Many2One(
         'stock.location', 'Location', required=True,
         domain=[('type', '=', 'warehouse')], states={
@@ -42,41 +39,38 @@ class Forecast(Workflow, ModelSQL, ModelView):
         depends=['state'])
     destination = fields.Many2One(
         'stock.location', 'Destination', required=True,
-        domain=[('type', 'in', ['customer', 'production'])], states=STATES,
-        depends=DEPENDS)
-    from_date = fields.Date('From Date', required=True, states=STATES,
-        depends=DEPENDS)
-    to_date = fields.Date('To Date', required=True, states=STATES,
-        depends=DEPENDS)
+        domain=[('type', 'in', ['customer', 'production'])], states=_states,
+        depends=_depends)
+    from_date = fields.Date(
+        "From Date", required=True,
+        domain=[('from_date', '<=', Eval('to_date'))],
+        states=_states, depends=_depends + ['to_date'])
+    to_date = fields.Date(
+        "To Date", required=True,
+        domain=[('to_date', '>=', Eval('from_date'))],
+        states=_states, depends=_depends + ['from_date'])
     lines = fields.One2Many(
-        'stock.forecast.line', 'forecast', 'Lines', states=STATES,
-        depends=DEPENDS)
+        'stock.forecast.line', 'forecast', 'Lines', states=_states,
+        depends=_depends)
     company = fields.Many2One(
         'company.company', 'Company', required=True, states={
             'readonly': Or(Not(Equal(Eval('state'), 'draft')),
                 Bool(Eval('lines', [0]))),
             },
         depends=['state'])
-    state = fields.Selection(
-        FORECAST_STATES, 'State', readonly=True, select=True)
+    state = fields.Selection([
+            ('draft', "Draft"),
+            ('done', "Done"),
+            ('cancel', "Cancel"),
+            ], "State", readonly=True, select=True)
     active = fields.Function(fields.Boolean('Active'),
         'get_active', searcher='search_active')
 
+    del _states, _depends
+
     @classmethod
     def __setup__(cls):
         super(Forecast, cls).__setup__()
-        cls.create_date.select = True
-        t = cls.__table__()
-        cls._sql_constraints += [
-            ('check_from_to_date', Check(t, t.to_date >= t.from_date),
-                '"To Date" must be greater than "From Date"'),
-            ]
-        cls._error_messages.update({
-                'date_overlap': ('Forecast "%(first)s" overlaps with dates '
-                    'of forecast "%(second)s" in the same location.'),
-                'delete_cancel': ('Forecast "%s" must be cancelled before '
-                    'deletion.'),
-                })
         cls._order.insert(0, ('from_date', 'DESC'))
         cls._order.insert(1, ('warehouse', 'ASC'))
         cls._transitions |= set((
@@ -109,7 +103,12 @@ class Forecast(Workflow, ModelSQL, ModelView):
     def __register__(cls, module_name):
         super(Forecast, cls).__register__(module_name)
 
+        # Add index on create_date
         table = cls.__table_handler__(module_name)
+        table.index_action('create_date', action='add')
+
+        # Migration from 5.0: remove check_from_to_date
+        table.drop_constraint('check_from_to_date')
 
     @staticmethod
     def default_state():
@@ -188,10 +187,10 @@ class Forecast(Workflow, ModelSQL, ModelView):
         forecast_id = cursor.fetchone()
         if forecast_id:
             second = self.__class__(forecast_id[0])
-            self.raise_user_error('date_overlap', {
-                    'first': self.rec_name,
-                    'second': second.rec_name,
-                    })
+            raise ForecastValidationError(
+                gettext('stock_forecast.msg_forecast_date_overlap',
+                    first=self.rec_name,
+                    second=second.rec_name))
 
     @classmethod
     def delete(cls, forecasts):
@@ -199,7 +198,9 @@ class Forecast(Workflow, ModelSQL, ModelView):
         cls.cancel(forecasts)
         for forecast in forecasts:
             if forecast.state != 'cancel':
-                cls.raise_user_error('delete_cancel', forecast.rec_name)
+                raise AccessError(
+                    gettext('stock_forecast.msg_forecast_delete_cancel',
+                        forecast=forecast.rec_name))
         super(Forecast, cls).delete(forecasts)
 
     @classmethod
@@ -272,11 +273,14 @@ class ForecastLine(ModelSQL, ModelView):
         depends=['product', 'product_uom_category'] + _depends)
     unit_digits = fields.Function(fields.Integer('Unit Digits'),
             'get_unit_digits')
-    quantity = fields.Float('Quantity', digits=(16, Eval('unit_digits', 2)),
-        required=True, states=_states, depends=['unit_digits'] + _depends)
-    minimal_quantity = fields.Float('Minimal Qty',
-        digits=(16, Eval('unit_digits', 2)), required=True,
+    quantity = fields.Float(
+        "Quantity", digits=(16, Eval('unit_digits', 2)), required=True,
+        domain=[('quantity', '>=', 0)],
         states=_states, depends=['unit_digits'] + _depends)
+    minimal_quantity = fields.Float(
+        "Minimal Qty", digits=(16, Eval('unit_digits', 2)), required=True,
+        domain=[('minimal_quantity', '<=', Eval('quantity'))],
+        states=_states, depends=['unit_digits', 'quantity'] + _depends)
     moves = fields.Many2Many('stock.forecast.line-stock.move',
         'line', 'move', 'Moves', readonly=True)
     forecast = fields.Many2One(
@@ -287,7 +291,7 @@ class ForecastLine(ModelSQL, ModelView):
             },
         depends=['forecast_state'])
     forecast_state = fields.Function(
-        fields.Selection(FORECAST_STATES, 'Forecast State'),
+        fields.Selection('get_forecast_states', 'Forecast State'),
         'on_change_with_forecast_state')
     quantity_executed = fields.Function(fields.Float('Quantity Executed',
             digits=(16, Eval('unit_digits', 2)), depends=['unit_digits']),
@@ -300,15 +304,20 @@ class ForecastLine(ModelSQL, ModelView):
         super(ForecastLine, cls).__setup__()
         t = cls.__table__()
         cls._sql_constraints += [
-            ('check_line_qty_pos', Check(t, t.quantity >= 0),
-                'Line quantity must be positive'),
-            ('check_line_minimal_qty',
-                Check(t, t.quantity >= t.minimal_quantity),
-                'Line quantity must be greater than the minimal quantity'),
             ('forecast_product_uniq', Unique(t, t.forecast, t.product),
-                'Product must be unique by forecast'),
+                'stock_forecast.msg_forecast_line_product_unique'),
             ]
 
+    @classmethod
+    def __register__(cls, module_name):
+        super().__register__(module_name)
+
+        table_h = cls.__table_handler__(module_name)
+
+        # Migration from 5.0: remove check on quantity
+        table_h.drop_constraint('check_line_qty_pos')
+        table_h.drop_constraint('check_line_minimal_qty')
+
     @staticmethod
     def default_unit_digits():
         return 2
@@ -338,6 +347,12 @@ class ForecastLine(ModelSQL, ModelView):
     def get_unit_digits(self, name):
         return self.product.default_uom.digits
 
+    @classmethod
+    def get_forecast_states(cls):
+        pool = Pool()
+        Forecast = pool.get('stock.forecast')
+        return Forecast.fields_get(['state'])['state']['selection']
+
     @fields.depends('forecast', '_parent_forecast.state')
     def on_change_with_forecast_state(self, name=None):
         if self.forecast:
@@ -366,7 +381,9 @@ class ForecastLine(ModelSQL, ModelView):
         line_move = LineMove.__table__()
 
         result = dict((x.id, 0) for x in lines)
-        key = lambda line: line.forecast.id
+
+        def key(line):
+            return line.forecast.id
         lines.sort(key=key)
         for forecast_id, lines in itertools.groupby(lines, key):
             forecast = Forecast(forecast_id)
@@ -500,19 +517,20 @@ class ForecastLineMove(ModelSQL):
 class ForecastCompleteAsk(ModelView):
     'Complete Forecast'
     __name__ = 'stock.forecast.complete.ask'
-    from_date = fields.Date('From Date', required=True)
-    to_date = fields.Date('To Date', required=True)
+    from_date = fields.Date(
+        "From Date", required=True,
+        domain=[('from_date', '<', Eval('to_date'))],
+        depends=['to_date'])
+    to_date = fields.Date(
+        "To Date", required=True,
+        domain=[('to_date', '>', Eval('from_date'))],
+        depends=['from_date'])
 
 
 class ForecastCompleteChoose(ModelView):
     'Complete Forecast'
     __name__ = 'stock.forecast.complete.choose'
-    products = fields.Many2Many(
-        'product.product', None, None, "Products",
-        domain=[
-            ('type', '=', 'goods'),
-            ('consumable', '=', False),
-            ])
+    products = fields.Many2Many('product.product', None, None, 'Products')
 
 
 class ForecastComplete(Wizard):
@@ -533,14 +551,6 @@ class ForecastComplete(Wizard):
             ])
     complete = StateTransition()
 
-    @classmethod
-    def __setup__(cls):
-        super(ForecastComplete, cls).__setup__()
-        cls._error_messages.update({
-                'from_to_date': (
-                    '"From Date" should be smaller than "To Date".'),
-                })
-
     def default_ask(self, fields):
         """
         Forecast dates shifted by one year.
@@ -558,8 +568,6 @@ class ForecastComplete(Wizard):
         Forecast = pool.get('stock.forecast')
         Product = pool.get('product.product')
         forecast = Forecast(Transaction().context['active_id'])
-        if self.ask.from_date > self.ask.to_date:
-            self.raise_user_error('from_to_date')
 
         with Transaction().set_context(
                 stock_destinations=[forecast.destination.id],
@@ -597,18 +605,21 @@ class ForecastComplete(Wizard):
             prod2line[forecast_line.product.id] = forecast_line
 
         pbl = self._get_product_quantity()
-        id2product = {p.id: p for p in Product.browse([x[1] for x in pbl])}
+        product_ids = [x[1] for x in pbl]
+        prod2uom = {}
+        for product in Product.browse(product_ids):
+            prod2uom[product.id] = product.default_uom.id
 
-        products = set(getattr(self.choose, 'products', {}))
+        if getattr(self.choose, 'products', None):
+            products = [x.id for x in self.choose.products]
+        else:
+            products = None
 
         to_save = []
         for key, qty in pbl.items():
-            _, product_id = key
-            product = id2product[product_id]
+            _, product = key
             if products and product not in products:
                 continue
-            if product.type != 'goods' or product.consumable:
-                continue
             if -qty <= 0:
                 continue
             if product in prod2line:
@@ -617,7 +628,7 @@ class ForecastComplete(Wizard):
                 line = ForecastLine()
             line.product = product
             line.quantity = -qty
-            line.uom = product.default_uom.id
+            line.uom = prod2uom[product]
             line.forecast = forecast
             line.minimal_quantity = min(1, -qty)
             to_save.append(line)
